from __future__ import annotations

import importlib
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from enum import EnumMeta
from typing import NamedTuple

from flask_principal import Permission

from vantage6.common import logger_name

from vantage6.backend.common.permission_models import RuleInterface

module_name = logger_name(__name__)
log = logging.getLogger(module_name)

RuleNeed = NamedTuple("RuleNeed", [("name", str), ("scope", str), ("operation", str)])


def get_attribute_name(
    operation: str, scope: str | None = None, shorten: bool = True
) -> str:
    """
    Get the attribute name for a rule.

    Parameters
    ----------
    operation: str
        Operation of the rule
    scope: str | None
        Scope of the rule
    shorten: bool
        Whether to shorten the attribute name in the format of 'v_glo' for view global.
        If False, the attribute name will be the full operation name.

    Returns
    -------
    str
        Attribute name for the rule
    """
    if shorten:
        attribute_name = operation[0].lower()
        if scope:
            attribute_name += f"_{scope[0:3].lower()}"
    else:
        attribute_name = operation
    return attribute_name


class RuleCollectionBase(ABC, dict):
    """
    Class that tracks a set of all rules for a certain resource name

    Parameters
    ----------
    name: str
        Name of the resource endpoint (e.g. node, organization, user)
    """

    def __init__(self, name):
        self.name = name

    def add(
        self, operation: str, scope: str | None = None, shorten: bool = True
    ) -> None:
        """
        Add a rule to the rule collection

        Parameters
        ----------
        scope: Scope
            Scope within which to apply the rule
        operation: Operation
            What operation the rule applies to
        """
        permission = Permission(RuleNeed(self.name, scope, operation))

        attribute_name = get_attribute_name(operation, scope, shorten)
        self.__setattr__(attribute_name, permission)


class PermissionManagerBase(ABC):
    """
    Loads the permissions and syncs rules in database with rules defined in
    the code
    """

    def __init__(
        self, resources_location: str, resources: list[str], default_roles: EnumMeta
    ) -> None:
        self.collections: dict[str, RuleCollectionBase] = {}
        self.default_roles = default_roles
        log.info("Loading permission system...")
        self.load_rules_from_resources(resources_location, resources)

    @abstractmethod
    def assign_rule_to_fixed_role(self, *args, **kwargs) -> None:
        """
        Attach a rule to a fixed role (not adjustable by users).

        Parameters
        ----------
        fixedrole: str
            Name of the fixed role that the rule should be added to
        resource: str
            Resource that the rule applies to
        operation: Operation
            Operation that the rule applies to
        """
        pass

    @abstractmethod
    def register_rule(self, *args, **kwargs) -> None:
        """
        Register a permission rule in the database.
        """
        pass

    @abstractmethod
    def check_user_rules(self, rules: list[RuleInterface]) -> dict | None:
        """
        Check if a user, node or container has all the `rules` in a list

        Parameters
        ----------
        rules: list[RuleInterface]
            List of rules that user is checked to have

        Returns
        -------
        dict | None
            Dict with a message which rule is missing, else None
        """
        pass

    @abstractmethod
    def get_new_collection(self, name: str) -> RuleCollectionBase:
        """
        Initialize and return a new RuleCollection.
        Parameters
        ----------
        name: str
            Name of the collection

        Returns
        -------
        RuleCollectionBase
            New RuleCollection
        """
        pass

    def load_rules_from_resources(
        self, resources_location: str, resources: list[str]
    ) -> None:
        """
        Collect all permission rules from all registered API resources.

        Parameters
        ----------
        resources_location: str
            Name of the module where to load the resources from (e.g. vantage6.server.resource).

        resources: list[str]
            List of the resources to load.
        """
        for res in resources:
            module = importlib.import_module(f"{resources_location}.{res}")
            try:
                module.permissions(self)
            except Exception:
                module_name = module.__name__.split(".")[-1]
                log.debug(
                    "Resource '%s' contains no or invalid permissions", module_name
                )

    def assign_rule_to_root(self, *args, **kwargs) -> None:
        """
        Assign a rule to the root role.

        Parameters
        ----------
        resource: str
            Resource that the rule applies to
        operation: Operation
            Operation that the rule applies to
        scope: Scope
            Scope that the rule applies to
        """

        self.assign_rule_to_fixed_role(self.default_roles.ROOT.value, *args, **kwargs)

    def appender(self, name: str) -> Callable:
        """
        Add a module's rules to the rule collection

        Parameters
        ----------
        name: str
            The name of the module whose rules are to be registered

        Returns
        -------
        Callable
            A Callable ``register_rule`` function
        """
        # make sure collection exists
        self.collection(name)
        return lambda *args, **kwargs: self.register_rule(name, *args, **kwargs)

    def collection(self, name: str) -> RuleCollectionBase:
        """
        Get a RuleCollection object. If it doesn't exist yet, it will be
        created.

        Parameters
        ----------
        name: str
            Name of the module whose RuleCollection is to be obtained or
            created

        Returns
        -------
        RuleCollectionBase
            The collection of rules belonging to the module name
        """
        if self._collection_exists(name):
            return self.collections[name]
        else:
            self.collections[name] = self.get_new_collection(name)
            return self.collections[name]

    def _collection_exists(self, name: str) -> bool:
        """
        Check if a module's rule collection is defined

        Parameters
        ----------
        name: str
            Name of the module to be checked

        Returns
        -------
        bool:
            True if RuleCollection is defined for module, else False
        """
        return name in self.collections

    def __getattr__(self, name: str) -> RuleCollectionBase:
        try:
            collection = self.collections[name]
            return collection
        except Exception as e:
            log.critical(f"Missing permission collection! {name}")
            raise e



import datetime
import enum

import sqlalchemy as sql
from sqlalchemy.ext.declarative import DeclarativeMeta

from vantage6.common import STRING_ENCODING
from vantage6.common.serialization import log


def jsonable(value: list[DeclarativeMeta] | DeclarativeMeta) -> list | dict:
    """
    Convert a (list of) SQLAlchemy instance(s) to native Python objects.

    Parameters
    ----------
    value : list[DeclarativeMeta] | DeclarativeMeta
        A single SQLAlchemy instance or a list of SQLAlchemy instances

    Returns
    -------
    list | dict
        A single Python object or a list of Python objects

    Raises
    ------
    Exception
        If the value is not an instance of db.Base or a list of db.Base
    """
    if isinstance(value, list):
        return [jsonable(i) for i in value]

    elif isinstance(value, DeclarativeMeta):
        log.debug(f"preparing={value}")
        retval = dict()
        mapper = sql.inspect(value.__class__)

        columns = [
            c.key for c in mapper.columns if c.key not in value._hidden_attributes
        ]

        for column in columns:
            column_value = getattr(value, column)

            if isinstance(column_value, enum.Enum):
                column_value = column_value.value
            elif isinstance(column_value, datetime.datetime):
                column_value = column_value.isoformat()
            elif isinstance(column_value, bytes):
                log.debug("decoding bytes!")
                column_value = column_value.decode(STRING_ENCODING)

            retval[column] = column_value

        return retval

    # FIXME: does it make sense to raise an exception or should base types
    #        (or other JSON-serializable types) just be returned as-is?
    raise Exception("value should be instance of db.Base or list!")



import logging
from typing import Any, Type

from prometheus_client import Gauge, start_http_server

log = logging.getLogger(__name__)


class Metric:
    """
    Represents a single metric with its name, expected type, and description.
    """

    def __init__(self, name: str, type_: Type[Any], description: str):
        self.name = name
        self.type = type_
        self.description = description


class Metrics:
    """
    Class to manage all system metrics and their Prometheus gauges.
    """

    def __init__(self, labels: list[str]):
        """
        Initialize the Metrics class and create Prometheus gauges.

        Parameters
        ----------
        labels: list[str]
            List of labels to be used for all metrics.
        """
        self.labels = labels
        self.gauges = {}
        self._initialize_metrics()

    def _initialize_metrics(self) -> None:
        """
        Define all system metrics and create their corresponding Prometheus gauges.
        """
        metrics = [
            Metric("cpu_percent", float, "CPU usage percentage"),
            Metric("memory_percent", float, "Memory usage percentage"),
            Metric(
                "num_algorithm_containers",
                int,
                "Number of running algorithm containers",
            ),
            Metric("cpu_count", int, "Number of CPUs"),
            Metric("memory_total", int, "Total memory"),
            Metric("memory_available", int, "Available memory"),
            Metric("gpu_count", int, "Number of GPUs"),
            Metric("gpu_load", float, "GPU load"),
            Metric("gpu_memory_used", int, "GPU memory used"),
            Metric("gpu_memory_free", int, "GPU memory free"),
            Metric("gpu_temperature", float, "GPU temperature"),
        ]

        for metric in metrics:
            if "gpu" in metric.name:
                self.gauges[metric.name] = Gauge(
                    metric.name, metric.description, labelnames=self.labels + ["gpu_id"]
                )
            else:
                self.gauges[metric.name] = Gauge(
                    metric.name, metric.description, labelnames=self.labels
                )

    def set_metric(self, metric_name: str, value: Any, labels: dict) -> None:
        """
        Set the value of a metric, handling GPU-specific logic if necessary.

        Parameters
        ----------
        metric_name: str
            The name of the metric to set.
        value: Any
            The value to set for the metric.
        labels: dict
            A dictionary of labels to apply to the metric.
        """
        if metric_name not in self.gauges:
            raise ValueError(f"Metric '{metric_name}' does not exist.")

        gauge = self.gauges[metric_name]

        if "gpu" in metric_name:
            # GPU metrics must be lists
            if not isinstance(value, list):
                raise ValueError(
                    f"Expected a list for GPU metric '{metric_name}', got {type(value).__name__}."
                )
            for gpu_id, gpu_value in enumerate(value):
                gauge.labels(**labels, gpu_id=gpu_id).set(gpu_value)
        else:
            # Non-GPU metrics
            gauge.labels(**labels).set(value)

    def get_gauge(self, metric_name: str) -> Gauge:
        """
        Retrieve the Prometheus gauge for a given metric.

        Parameters
        ----------
        metric_name: str
            The name of the metric.

        Returns
        -------
        Gauge
            The Prometheus gauge for the metric.
        """
        if metric_name not in self.gauges:
            raise ValueError(f"Metric '{metric_name}' does not exist.")
        return self.gauges[metric_name]


def start_prometheus_exporter(port: int = 9100) -> None:
    """
    Start the Prometheus exporter to expose metrics.
    """
    log.info("Initializing Prometheus exporter...")
    try:
        start_http_server(port)
        log.info("Prometheus exporter started on port %s", port)
    except Exception as e:
        log.error("Failed to start Prometheus exporter: %s", e)



import logging

from flask import url_for
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema
from sqlalchemy.ext.declarative import DeclarativeMeta

from vantage6.common import logger_name

from vantage6.backend.common.resource.pagination import Pagination

log = logging.getLogger(logger_name(__name__))


def create_one_to_many_link(obj: DeclarativeMeta, link_to: str, link_from: str) -> str:
    """
    Create an API link to get objects related to a given object.

    Parameters
    ----------
    obj : DeclarativeMeta
        Object to which the link is created
    link_to : str
        Name of the resource to which the link is created
    link_from : str
        Name of the resource from which the link is created

    Returns
    -------
    str
        API link

    Examples
    --------
    >>> create_one_to_many_link(obj, "node", "organization_id")
    "/api/node?organization_id=<obj.id>"
    """
    endpoint = link_to + "_without_id"
    values = {link_from: obj.id}
    return url_for(endpoint, **values)


class BaseHATEOASModelSchema(SQLAlchemyAutoSchema):
    """Base class for generating HATEOAS links for SQLAlchemy resources."""

    api = None

    def create_hateoas(
        self, name: str, obj: DeclarativeMeta, endpoint: str = None
    ) -> dict | None:
        """
        Create a HATEOAS link to a related object.

        Parameters
        ----------
        name : str
            Name of the related resource
        obj : Derivative of DeclarativeMeta class
            SQLAlchemy resource to which the link is created
        endpoint : str, optional
            Name of the endpoint to which the link is created, by default None.
            If None, the endpoint is assumed to be the same as the name of the
            related resource.

        Returns
        -------
        dict | None
            HATEOAS link to the related object, or None if the related object
            does not exist.
        """
        # get the related object
        elem = getattr(obj, name)

        # create the link
        endpoint = endpoint if endpoint else name
        if elem:
            return self._hateoas_link_from_related_resource(elem, endpoint)
        else:
            return None

    def _hateoas_link_from_related_resource(
        self, elem: DeclarativeMeta, name: str
    ) -> dict:
        """
        Construct a HATEOAS link from a related object.

        Parameters
        ----------
        obj : Derivative of DeclarativeMeta class
            SQLAlchemy resource for which the link is created
        name : str
            Name of the resource

        Returns
        -------
        dict
            HATEOAS link to the related object
        """
        _id = elem.id
        endpoint = name + "_with_id"
        if self.api:
            if not self.api.owns_endpoint(endpoint):
                Exception(f"Make sure {endpoint} exists!")

            verbs = list(self.api.app.url_map._rules_by_endpoint[endpoint][0].methods)
            verbs.remove("HEAD")
            verbs.remove("OPTIONS")
            url = url_for(endpoint, id=_id)
            return {"id": _id, "link": url, "methods": verbs}
        else:
            log.error("No API found?")

    def meta_dump(self, pagination: Pagination) -> dict:
        """
        Dump paginated database resources to a dictionary that has links
        to additional pages.

        Parameters
        ----------
        pagination : Pagination
            Paginated database resources

        Returns
        -------
        dict
            Dictionary with paginated database resources and links to
            additional pages.
        """
        data = self.dump(pagination.page.items, many=True)
        return {"data": data, "links": pagination.metadata_links}



import logging
from typing import Any, List

from requests import Session
from sqlalchemy import or_

from vantage6.common import logger_name

from vantage6.backend.common.permission import PermissionManagerBase
from vantage6.backend.common.resource.error_handling import (
    BadRequestError,
    NotFoundError,
)

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def validate_request_body(schema, data, partial=False):
    errors = schema.validate(data, partial=partial)
    if errors:
        raise BadRequestError(errors)
    return None


def get_rules(data: Any, db: Session) -> List:
    rules = []
    if data["rules"]:
        for rule_id in data["rules"]:
            rule = db.Rule.get(rule_id)
            if not rule:
                raise NotFoundError(f"Rule with id {rule_id} not found")
            rules.append(rule)
    return rules


def filter_by_attribute(db, params: list, query, args):
    for param in params:
        filters = args.getlist(param)
        if filters:
            query = query.filter(
                or_(*[getattr(db.Role, param).like(f) for f in filters])
            )
    return query


def validate_user_exists(db, user_id):
    user = db.User.get(user_id)
    if not user:
        raise BadRequestError(f"User with id={user_id} does not exist!")
    return user


def apply_user_filter(db, query, user_id):
    return query.join(db.Permission).join(db.User).filter(db.User.id == user_id)


def get_role(db, role_id):
    role = db.Role.get(role_id)
    if not role:
        raise NotFoundError(f"Role with id={role_id} not found.")
    return role


def check_default_role(role, default_roles):
    if role.name in default_roles:
        raise BadRequestError(
            f"Role {role.name} is a default role and cannot be edited or deleted."
        )


def get_rule(db, rule_id):
    rule = db.Rule.get(rule_id)
    if not rule:
        raise NotFoundError(f"Rule with id={rule_id} not found.")
    return rule


def get_rules_from_ids(rule_ids, db):
    rules = []
    for rule_id in rule_ids:
        rule = get_rule(db, rule_id)
        rules.append(rule)
    return rules


def update_role(role, data, db, permissions: PermissionManagerBase):
    if "name" in data:
        role.name = data["name"]
    if "description" in data:
        role.description = data["description"]
    if "rules" in data:
        rules = get_rules_from_ids(data["rules"], db)
        permissions.check_user_rules(rules)
        role.rules = rules
    return role


def can_delete_dependents(role, delete_dependents):
    if role.users:
        if delete_dependents:
            log.warn(
                f"Role {id} will be deleted even though it was assigned to users. This may result in missing permissions."
            )
        else:
            raise BadRequestError(
                f"Role {role.name} is assigned to users and cannot be deleted."
            )



from vantage6.common.enum import StrEnumBase

# pagination settings
DEFAULT_PAGE = 1
DEFAULT_PAGE_SIZE = 10

# environment variable name for host URI
HOST_URI_ENV = "HOST_URI_ENV_VAR"

# database attempts setting
MAX_NUMBER_OF_ATTEMPTS = 10
RETRY_DELAY_IN_SECONDS = 30

# default support email address to make users aware of
DEFAULT_SUPPORT_EMAIL_ADDRESS = "support@vantage6.ai"

# default email address used in 'from' header
DEFAULT_EMAIL_FROM_ADDRESS = "noreply@vantage6.ai"


class RequiredServerEnvVars(StrEnumBase):
    """Enum containing the required server environment variables"""

    KEYCLOAK_URL = "KEYCLOAK_URL"
    KEYCLOAK_REALM = "KEYCLOAK_REALM"
    KEYCLOAK_ADMIN_USERNAME = "KEYCLOAK_ADMIN_USERNAME"
    KEYCLOAK_ADMIN_PASSWORD = "KEYCLOAK_ADMIN_PASSWORD"
    KEYCLOAK_ADMIN_CLIENT = "KEYCLOAK_ADMIN_CLIENT"
    KEYCLOAK_ADMIN_CLIENT_SECRET = "KEYCLOAK_ADMIN_CLIENT_SECRET"



"""Common functionality for the vantage6 server and algorithm store."""

import importlib.metadata
import json
import logging
import os
import traceback
from abc import abstractmethod
from http import HTTPStatus
from pathlib import Path
from types import ModuleType

import requests
from flask import Flask, Request, Response, make_response, request, send_from_directory
from flask_cors import CORS
from flask_cors.core import probably_regex
from flask_jwt_extended import JWTManager
from flask_marshmallow import Marshmallow
from flask_principal import Principal
from flask_restful import Api
from requests import Response as RequestsResponse
from werkzeug.exceptions import HTTPException

from vantage6.common import logger_name, validate_required_env_vars
from vantage6.common.globals import DEFAULT_API_PATH, LOCALHOST

from vantage6.cli.context.base_server import BaseServerContext

from vantage6.backend.common.auth import get_keycloak_id_for_user
from vantage6.backend.common.base import BaseDatabaseSessionManager, BaseModelBase
from vantage6.backend.common.globals import (
    DEFAULT_SUPPORT_EMAIL_ADDRESS,
    HOST_URI_ENV,
    RequiredServerEnvVars,
)
from vantage6.backend.common.jsonable import jsonable
from vantage6.backend.common.mail_service import MailService
from vantage6.backend.common.resource.output_schema import BaseHATEOASModelSchema

__version__ = importlib.metadata.version(__package__)

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


class Vantage6App:
    """Base class for all vantage6 server applications."""

    def __init__(self, ctx: BaseServerContext, server_module_name: str) -> None:
        """Initialize the vantage6 app."""
        self.ctx = ctx

        # validate that the required environment variables are set
        self.validate_required_env_vars()

        # initialize, configure Flask
        self.app = Flask(
            server_module_name,
            root_path=Path(__file__),
            template_folder=Path(__file__).parent / "templates",
            static_folder=Path(__file__).parent / "static",
        )
        self.debug: dict = self.ctx.config.get("debug", {})

        # configure the flask app
        self.configure_flask()

        # Setup SQLAlchemy and Marshmallow for marshalling/serializing
        self.ma = Marshmallow(self.app)

        # Setup the Flask-JWT-Extended extension (JWT: JSON Web Token)
        self.jwt = JWTManager(self.app)
        self.configure_jwt()

        # Setup Principal, granular API access manegement
        self.principal = Principal(self.app, use_sessions=False)

        # Enable cross-origin resource sharing. Note that Flask-CORS interprets
        # the origins as regular expressions.
        cors_allowed_origins = self.ctx.config.get("cors_allowed_origins", "*")
        self._warn_if_cors_regex(cors_allowed_origins)
        self.cors = CORS(
            self.app,
            resources={r"/*": {"origins": cors_allowed_origins}},
        )

        # Setup the Flask-Mail client
        self.mail = MailService(self.app)

        # Api - REST JSON-rpc
        self.api = Api(self.app)
        self.configure_api()

        # set environment variable for dev environment
        host_uri = self.ctx.config.get("dev", {}).get("host_uri")
        if host_uri:
            os.environ[HOST_URI_ENV] = host_uri

        # set the server version
        self.__version__ = __version__

    @abstractmethod
    def configure_flask(self) -> None:
        """Configure the flask app."""

    @abstractmethod
    def configure_jwt(self) -> None:
        """Configure the JWT extension."""

    @abstractmethod
    def configure_api(self) -> None:
        """Configure the API."""

    @abstractmethod
    def load_resources(self) -> None:
        """Load the resources."""

    def validate_required_env_vars(self) -> None:
        """Validate that the required environment variables are set."""
        validate_required_env_vars(RequiredServerEnvVars)

    def _get_keycloak_public_key(self) -> str:
        """Get the public key for the keycloak server."""
        response = requests.get(
            f"{os.environ.get(RequiredServerEnvVars.KEYCLOAK_URL.value)}/realms"
            f"/{os.environ.get(RequiredServerEnvVars.KEYCLOAK_REALM.value)}",
            timeout=100,
        )
        key = response.json()["public_key"]
        return f"-----BEGIN PUBLIC KEY-----\n{key}\n-----END PUBLIC KEY-----"

    def _configure_flask_base(
        self, database_session_manager: type["BaseDatabaseSessionManager"]
    ) -> None:
        """Configure the flask app."""
        # let us handle exceptions
        self.app.config["PROPAGATE_EXCEPTIONS"] = True

        # set JWT algorithms that keycloak uses
        self.app.config["JWT_ALGORITHM"] = "RS256"
        self.app.config["JWT_DECODE_ALGORITHMS"] = ["RS256"]
        # Leeway is provided for the token IAT to prevent errors that token is not yet
        # valid, which can happen if server times are drifting slightly.
        self.app.config["JWT_DECODE_LEEWAY"] = self.ctx.config.get(
            "jwt_decode_leeway", 10
        )
        self.app.config["JWT_PUBLIC_KEY"] = self._get_keycloak_public_key()
        self.app.config.setdefault("JWT_TOKEN_LOCATION", ["headers"])

        # Mail settings
        mail_config = self.ctx.config.get("smtp", {})
        self.app.config["MAIL_PORT"] = mail_config.get("port", 1025)
        self.app.config["MAIL_SERVER"] = mail_config.get("server", LOCALHOST)
        self.app.config["MAIL_USERNAME"] = mail_config.get(
            "username", DEFAULT_SUPPORT_EMAIL_ADDRESS
        )
        self.app.config["MAIL_PASSWORD"] = mail_config.get("password", "")
        self.app.config["MAIL_USE_TLS"] = mail_config.get("MAIL_USE_TLS", True)
        self.app.config["MAIL_USE_SSL"] = mail_config.get("MAIL_USE_SSL", False)
        debug_mode = self.debug.get("flask", False)
        if debug_mode:
            log.debug("Flask debug mode enabled")
        self.app.debug = debug_mode

        def _get_request_path(request_: Request) -> str:
            """
            Return request extension of request URL, e.g.
            http://localhost:7601/api/task/1 -> api/task/1

            Parameters
            ----------
            request: Request
                Flask request object

            Returns
            -------
            string:
                The endpoint path of the request
            """
            return request_.url.replace(request_.url_root, "")

        # before request
        @self.app.before_request
        def do_before_request():
            """Before every flask request method."""
            # Add log message before each request
            log.debug(
                "Received request: %s %s", request.method, _get_request_path(request)
            )

            # This will obtain a (scoped) db session from the session factory
            # that is linked to the flask request global `g`. In every endpoint
            # we then can access the database by using this session. We ensure
            # that the session is removed (and uncommited changes are rolled
            # back) at the end of every request.
            database_session_manager.new_session()

        @self.app.after_request
        def remove_db_session(response):
            """After every flask request.

            This will close the database session created by the
            `before_request`.
            """
            database_session_manager.clear_session()
            return response

        @self.app.errorhandler(HTTPException)
        def error_remove_db_session(error: HTTPException):
            """In case an HTTP-exception occurs during the request.

            It is important to close the db session to avoid having dangling
            sessions.
            """
            if error.code == 404:
                log.debug("404 error for route '%s'", _get_request_path(request))
            else:
                log.warning("HTTP Exception occured during request")
                log.debug("Details exception: %s", traceback.format_exc())
            database_session_manager.clear_session()
            return error.get_response()

        @self.app.errorhandler(Exception)
        def error2_remove_db_session(_: Exception):
            """In case an exception occurs during the request.

            It is important to close the db session to avoid having dangling
            sessions.
            """
            log.exception("Exception occured during request")
            database_session_manager.clear_session()
            return {
                "msg": "An unexpected error occurred on the server!"
            }, HTTPStatus.INTERNAL_SERVER_ERROR

        @self.app.route("/robots.txt")
        def static_from_root():
            return send_from_directory(self.app.static_folder, request.path[1:])

    def _configure_api_base(
        self,
        hateoas_schema: type["BaseHATEOASModelSchema"],
        base_db_model: type["BaseModelBase"],
    ) -> None:
        """Configure the API."""

        hateoas_schema.api = self.api

        # whatever you get try to json it
        @self.api.representation("application/json")
        # pylint: disable=unused-argument
        def output_json(
            data: base_db_model | list[base_db_model],
            code: HTTPStatus,
            headers: dict = None,
        ) -> Response:
            """
            Return jsonified data for request responses.

            Parameters
            ----------
            data: db.Base | list[db.Base]
                The data to be jsonified
            code: HTTPStatus
                The HTTP status code of the response
            headers: dict
                Additional headers to be added to the response
            """

            if isinstance(data, base_db_model):
                data = jsonable(data)
            elif (
                isinstance(data, list)
                and len(data)
                and isinstance(data[0], base_db_model)
            ):
                data = jsonable(data)
            # Don't attempt json conversion if it's already a response.
            # Used for streaming endpoints where responses cannot be jsonified.
            elif isinstance(data, RequestsResponse):
                resp = make_response(data, code)
                return resp

            resp = make_response(json.dumps(data), code)
            resp.headers.extend(headers or {})
            return resp

    @staticmethod
    def _add_default_roles(default_roles: list, db: ModuleType) -> None:
        for role in default_roles:
            if not db.Role.get_by_name(role["name"]):
                log.warning("Creating new default role %s", role["name"])
                new_role = db.Role(
                    name=role["name"],
                    description=role["description"],
                    rules=role["rules"],
                    is_default_role=True,
                )
                new_role.save()
            else:
                current_role = db.Role.get_by_name(role["name"])
                # check that the rules are the same. Use set() to compare without order
                if set(current_role.rules) != set(role["rules"]):
                    log.warning("Updating default role %s with new rules", role["name"])
                    current_role.rules = role["rules"]
                    current_role.save()

    def _add_keycloak_id_to_super_user(self, super_user: BaseModelBase) -> None:
        try:
            super_user.keycloak_id = get_keycloak_id_for_user(super_user.username)
            super_user.save()
        except Exception as exc:
            log.error(
                "Could not get keycloak ID for super user %s", super_user.username
            )
            log.error("This means that you cannot login as this user")
            log.exception(exc)

    @staticmethod
    def _warn_if_cors_regex(origins: str | list[str]) -> None:
        """
        Give a warning if CORS origins are regular expressions. This will not work
        properly for socket events (Flask-SocketIO checks for string equality and does
        not use regex).

        Note that we are using the `probably_regex` function from Flask-CORS to check
        if the origins are probably regular expressions - the Flask implementation for
        determining if it is a regex is a bit hacky (see
        https://github.com/corydolphin/flask-cors/blob/3.0.10/flask_cors/core.py#L275-L285)
        and Flask-CORS doesn't currently offer an opt out of regex's altogether.

        Parameters
        ----------
        origins: str | list[str]
            The origins to check
        """
        if isinstance(origins, str):
            origins = [origins]

        for origin in origins:
            if probably_regex(origin) and not origin == "*":
                log.warning(
                    "CORS origin '%s' is a regular expression. Socket events sent from "
                    "this origin will not be handled properly.",
                    origin,
                )


def get_server_url(
    config: dict, server_url_from_request: str | None = None
) -> str | None:
    """ "
    Get the server url from the request data, or from the configuration if it is
    not present in the request.

    Parameters
    ----------
    config : dict
        Server configuration
    server_url_from_request : str | None
        Server url from the request data.

    Returns
    -------
    str | None
        The server url
    """
    if server_url_from_request:
        return server_url_from_request
    server_url = config.get("server_url")
    # make sure that the server url ends with the api path
    api_path = config.get("api_path", DEFAULT_API_PATH)
    if server_url and not server_url.endswith(api_path):
        server_url = server_url + api_path
    return server_url



import logging
import os
from dataclasses import dataclass

from keycloak import KeycloakAdmin, KeycloakOpenID

from vantage6.backend.common.globals import RequiredServerEnvVars
from vantage6.backend.common.resource.error_handling import BadRequestError

log = logging.getLogger(__name__)


@dataclass
class KeycloakServiceAccount:
    """Details about a service account in Keycloak"""

    client_id: str
    client_secret: str
    user_id: str


def _get_admin_token():
    """Get a token for the admin client"""
    keycloak_openid = KeycloakOpenID(
        server_url=os.environ.get(RequiredServerEnvVars.KEYCLOAK_URL.value),
        client_id=os.environ.get(RequiredServerEnvVars.KEYCLOAK_ADMIN_CLIENT.value),
        realm_name=os.environ.get(RequiredServerEnvVars.KEYCLOAK_REALM.value),
        client_secret_key=os.environ.get(
            RequiredServerEnvVars.KEYCLOAK_ADMIN_CLIENT_SECRET.value
        ),
    )

    # Get token using client credentials (service account) flow
    return keycloak_openid.token(grant_type="client_credentials")


def get_keycloak_admin_client():
    """
    Get a KeycloakAdmin client
    """
    token = _get_admin_token()

    # Create KeycloakAdmin with the service account token
    keycloak_admin = KeycloakAdmin(
        server_url=os.environ.get(RequiredServerEnvVars.KEYCLOAK_URL.value),
        realm_name=os.environ.get(RequiredServerEnvVars.KEYCLOAK_REALM.value),
        token=token,
        verify=True,
    )

    return keycloak_admin


def get_keycloak_id_for_user(username: str):
    """
    Get the keycloak id for a user
    """
    try:
        keycloak_admin: KeycloakAdmin = get_keycloak_admin_client()
        keycloak_id = keycloak_admin.get_user_id(username)
    except Exception as exc:
        log.exception(exc)
        raise BadRequestError("Could not retrieve user from Keycloak") from exc

    if keycloak_id is None:
        raise BadRequestError("User does not exist in Keycloak")
    return keycloak_id


def create_service_account_in_keycloak(
    client_name: str, is_node: bool = True
) -> KeycloakServiceAccount:
    """
    Create a service account in Keycloak
    """
    try:
        keycloak_admin: KeycloakAdmin = get_keycloak_admin_client()
        client_id = keycloak_admin.create_client(
            {
                "clientId": client_name,
                "publicClient": False,
                "enabled": True,
                "serviceAccountsEnabled": True,
                "standardFlowEnabled": False,
                "protocolMappers": [
                    {
                        "name": "vantage6_client_type",
                        "protocol": "openid-connect",
                        "protocolMapper": "oidc-hardcoded-claim-mapper",
                        "config": {
                            "claim.name": "vantage6_client_type",
                            "claim.value": "node" if is_node else "user",
                            "access.token.claim": True,
                        },
                    }
                ],
            }
        )
        user_id = keycloak_admin.get_user_id(f"service-account-{client_name}")
        secret = keycloak_admin.get_client_secrets(client_id)
    except Exception as exc:
        log.exception(exc)
        raise BadRequestError(
            f"Could not create service account '{client_name}' in Keycloak"
        ) from exc
    return KeycloakServiceAccount(client_id, secret["value"], user_id)


def get_service_account_in_keycloak(client_name: str) -> KeycloakServiceAccount:
    """
    Get a service account in Keycloak
    """
    keycloak_admin: KeycloakAdmin = get_keycloak_admin_client()
    client_id = keycloak_admin.get_client_id(client_name)
    user_id = keycloak_admin.get_user_id(f"service-account-{client_name}")
    secret = keycloak_admin.get_client_secrets(client_id)
    return KeycloakServiceAccount(client_id, secret["value"], user_id)


def delete_service_account_in_keycloak(client_id: str) -> None:
    """
    Delete a service account in Keycloak
    """
    try:
        keycloak_admin: KeycloakAdmin = get_keycloak_admin_client()
        keycloak_admin.delete_client(client_id)
    except Exception as exc:
        log.exception(exc)
        raise BadRequestError(
            f"Service account '{client_id}' could not be deleted from Keycloak"
        ) from exc



from __future__ import annotations

import logging
import math
from urllib.parse import urlencode

import flask
from flask import g
from sqlalchemy import desc, func, select
from sqlalchemy.orm.decl_api import DeclarativeMeta
from sqlalchemy.sql.selectable import Select

from vantage6.common import logger_name

from vantage6.backend.common.globals import DEFAULT_PAGE, DEFAULT_PAGE_SIZE

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


class Page:
    """
    Definition of a page of items return by the API.

    Parameters
    ----------
    items : list[DeclarativeMeta]
        List of database resources on this page
    page : int
        Current page number
    page_size : int
        Number of items per page
    total : int
        Total number of items

    Attributes
    ----------
    current_page : int
        Current page number
    items : list[DeclarativeMeta]
        List of resources on the current page
    previous_page : int
        Page number of the previous page
    next_page : int
        Page number of the next page
    has_previous : bool
        True if there is a previous page, False otherwise
    has_next : bool
        True if there is a next page, False otherwise
    total : int
        Total number of items
    pages : int
        Total number of pages
    """

    def __init__(
        self, items: list[DeclarativeMeta], page: int, page_size: int, total: int
    ) -> None:
        self.current_page = page
        self.items = items
        self.previous_page = None
        self.next_page = None
        self.has_previous = page > 1
        if self.has_previous:
            self.previous_page = page - 1
        previous_items = (page - 1) * page_size
        self.has_next = previous_items + len(items) < total
        if self.has_next:
            self.next_page = page + 1
        self.total = total
        self.pages = int(math.ceil(total / float(page_size)))


class Pagination:
    """
    Class that handles pagination of a query.

    Parameters
    ----------
    items : list[DeclarativeMeta]
        List of database resources to paginate
    page : int
        Current page number
    page_size : int
        Number of items per page
    total : int
        Total number of items
    request : flask.Request
        Request object

    Attributes
    ----------
    page : Page
        Page object
    request : flask.Request
        Request object
    """

    def __init__(
        self,
        items: list[DeclarativeMeta],
        page: int,
        page_size: int,
        total: int,
        request: flask.Request,
    ) -> None:
        self.page = Page(items, page, page_size, total)
        self.request = request

    @property
    def link_header(self) -> str:
        """
        Puts links to other pages in the response header.

        Returns
        -------
        str
            Link header
        """
        link_strs = [f"<{url}>; rel={rel}" for rel, url in self.metadata_links.items()]
        return ",".join(link_strs)

    @property
    def headers(self) -> dict:
        """
        Set the headers for the response.

        Returns
        -------
        dict
            Response headers
        """
        return {
            "total-count": self.page.total,
            "Link": self.link_header,
            # indicate that these headers are allowed to be exposed to scripts
            # running in a browser
            "access-control-expose-headers": "total-count, Link",
        }

    @property
    def metadata_links(self) -> dict:
        """
        Construct links to other pages.

        Returns
        -------
        dict
            Links to other pages
        """
        url = self.request.path
        args = self.request.args.copy()

        navs = [
            {"rel": "first", "page": 1},
            {"rel": "previous", "page": self.page.previous_page},
            {"rel": "self", "page": self.page.current_page},
            {"rel": "next", "page": self.page.next_page},
            {"rel": "last", "page": self.page.pages},
        ]

        links = {}
        for nav in navs:
            if nav["page"]:
                args["page"] = nav["page"]
                links[nav["rel"]] = f"{url}?{urlencode(args)}"

        return links

    @classmethod
    def from_query(
        cls,
        query: Select,
        request: flask.Request,
        resource_model: DeclarativeMeta,
        paginate: bool = True,
    ) -> Pagination:
        """
        Create a Pagination object from a query.

        Parameters
        ----------
        query : sqlalchemy.sql.selectable.Select
            Query to paginate
        request : flask.Request
            Request object
        resource_model : DeclarativeMeta
            SQLAlchemy model of the resource whose endpoint is being called
        paginate : bool
            Whether to paginate the query or not, default True

        Returns
        -------
        Pagination
            Pagination object
        """
        # We remove the ordering of the query since it doesn't matter for
        # getting a count and might have performance implications as discussed
        # on this Flask-SqlAlchemy issue
        # https://github.com/mitsuhiko/flask-sqlalchemy/issues/100
        # pylint: disable=not-callable
        count_query = select(func.count()).select_from(
            query.distinct().order_by(None).subquery()
        )
        total = g.session.scalar(count_query)

        # check if pagination is desired, else return all records
        if paginate:
            page_id = int(request.args.get("page", 1))
            per_page = int(request.args.get("per_page", 10))

            if page_id <= 0:
                raise AttributeError("page needs to be >= 1")
            if per_page <= 0:
                raise AttributeError("per_page needs to be >= 1")
        else:
            page_id = 1
            per_page = total or 1

        # FIXME BvB 2020-02-09 good error handling if sort is not a valid
        #  field
        if request.args.get("sort", False):
            query = cls._add_sorting(query, request.args.get("sort"), resource_model)

        items = g.session.scalars(
            query.distinct().limit(per_page).offset((page_id - 1) * per_page)
        ).all()

        return cls(items, page_id, per_page, total, request)

    @staticmethod
    def _get_page_id(request: flask.Request) -> int:
        """
        Get the page id from the request.

        Parameters
        ----------
        request : flask.Request
            Request object

        Returns
        -------
        int
            Page id

        Raises
        ------
        ValueError
            If the page id is not an integer or is less than 1
        """
        try:
            page_id = int(request.args.get("page", DEFAULT_PAGE))
        except ValueError:
            raise ValueError("The 'page' parameter should be an integer")
        if page_id <= 0:
            raise ValueError("The 'page' parameter should be >= 1")
        return page_id

    @staticmethod
    def _get_per_page(request: flask.Request) -> int:
        """
        Get the number of items per page from the request.

        Parameters
        ----------
        request : flask.Request
            Request object

        Returns
        -------
        int
            Number of items per page

        Raises
        ------
        ValueError
            If the number of items per page is not an integer or is less than 1
        """
        try:
            per_page = int(request.args.get("per_page", DEFAULT_PAGE_SIZE))
        except ValueError:
            raise ValueError("The 'per_page' parameter should be an integer")
        if per_page <= 0:
            raise ValueError("The 'per_page' parameter should be >= 1")
        return per_page

    @staticmethod
    def _add_sorting(
        query: Select, sort_string: str, resource_model: DeclarativeMeta
    ) -> Select:
        """
        Add sorting to a query.

        Parameters
        ----------
        query : sqlalchemy.sql.selectable.Select
            The query to add sorting to.
        sort : str
            The sorting to add. This can be a comma separated list of fields to
            sort on. The fields can be prefixed with a '-' to indicate a
            descending sort.
        resource_model : DeclarativeMeta
            SQLAlchemy model of the resource whose endpoint is being called
        """
        sort_list = sort_string.split(",")
        for sorter in sort_list:
            sorter = sorter.strip()
            sort_asc = True
            if sorter.startswith("-"):
                sort_asc = False
            if sorter.startswith("+") or sorter.startswith("-"):
                sorter = sorter[1:]
            if not hasattr(resource_model, sorter):
                raise AttributeError(
                    f"Cannot sort a '{resource_model.__name__}' by "
                    f"'{sorter}' because it does not exist"
                )
            sorter = getattr(resource_model, sorter)
            if sort_asc:
                query = query.order_by(sorter)
            else:
                query = query.order_by(desc(sorter))
        return query



from __future__ import annotations

from vantage6.common.context import AppContext
from vantage6.common.globals import APPNAME, InstanceType

from vantage6.cli.configuration_manager import TestingConfigurationManager


class TestContext(AppContext):
    """
    Server context for testing purposes.

    Note that this context is specific to the server: for nodes, there is a
    separate test context.
    """

    INST_CONFIG_MANAGER = TestingConfigurationManager
    LOGGING_ENABLED = False

    @classmethod
    def from_external_config_file(
        cls, package_folder: str, instance_type: InstanceType
    ) -> TestContext:
        """
        Create a context the unittest configuration file.

        Returns
        -------
        TestContext
            Context object
        """
        return super().from_external_config_file(
            cls.test_config_location(package_folder, instance_type), instance_type, True
        )

    @staticmethod
    def test_config_location(package_folder: str, instance_type: InstanceType) -> str:
        """
        Location of the unittest configuration file.

        Returns
        -------
        str
            Path to the unittest configuration file
        """
        if instance_type == InstanceType.SERVER:
            return (
                package_folder / APPNAME / "server" / "_data" / "unittest_config.yaml"
            )
        elif instance_type == InstanceType.ALGORITHM_STORE:
            return (
                package_folder
                / APPNAME
                / "algorithm"
                / "store"
                / "_data"
                / "unittest_config.yaml"
            )
